"""Instantiate Configuration class and set default values."""

import shutil
from pathlib import Path
from typing import Any

import tomllib
import typer
from loguru import logger
from rich import print as rprint

PATH_CONFIG_DEFAULT = Path(__file__).parent / "default_config.toml"


class Config:
    """Representation of a configuration file."""

    def __init__(
        self,
        config_path: Path | None = None,
        context: dict[str, Any] = {},
    ) -> None:
        """Initialize configuration file."""
        self.config_path = config_path.expanduser().resolve() if config_path else None
        self.context = context

        if not config_path or not self.config_path.exists():
            self._create_config()

        self.config = self._load_config()

    def __repr__(self) -> str:
        """Return string representation of Config.

        Returns:
            str: The string representation of the Config object.
        """
        return f"{self.config}"

    def _create_config(self) -> None:
        """Create a configuration file from the default when it does not exist.

        Raises:
            typer.Exit: If no configuration file is specified.
        """
        if self.config_path is None:
            logger.error("No configuration file specified")
            raise typer.Exit(code=1)

        self.config_path.parent.mkdir(parents=True, exist_ok=True)
        shutil.copy(PATH_CONFIG_DEFAULT, self.config_path)
        logger.success(f"Created default configuration file at {self.config_path}")
        logger.info(f"Please edit {self.config_path} before continuing")
        rprint(f"{self.config_path} created. Please edit before continuing.")
        raise typer.Exit(code=1)

    def _load_config(self) -> dict[str, Any]:
        """Load the configuration file.

        This method reads the configuration file specified by `self.config_path`
        and returns the parsed configuration as a dictionary. It also adds the
        context items to the configuration dictionary.

        Returns:
            A dictionary containing the parsed configuration.

        Raises:
            typer.Exit: If the configuration file is empty or malformed.
        """
        logger.trace(f"Loading configuration from {self.config_path}")
        with self.config_path.open("rb") as f:
            try:
                config = tomllib.load(f)
            except tomllib.TOMLDecodeError as e:
                logger.exception(f"Could not parse '{self.config_path}'")
                raise typer.Exit(code=1) from e

        if config == {}:
            logger.error(f"Configuration file '{self.config_path}' is empty or malformed")
            raise typer.Exit(code=1)

        # Add context to config
        for key, value in self.context.items():
            config[key] = value

        return config

    def get(
        self, key: str, default: str | bool | list[str] | None = None, pass_none: bool = False
    ) -> str | int | bool | list[str | int] | None:
        """Get a value from the configuration file.

        Args:
            key (str): The name of the config variable.
            default (str, optional): The default value to use if the key is not set. Defaults to None.
            pass_none (bool, optional): Whether to pass None if the key does not exist. Defaults to False.

        Returns:
            str | int | bool | list[str | int] | None: The value of the config variable.

        Raises:
            typer.Exit: If the config variable is not set and `pass_none` is False.
        """
        value = self.config.get(key, default)

        if value is None and not pass_none:
            logger.error(f"Config variable {key} is not set.")
            raise typer.Exit(code=1)

        if not value and not isinstance(value, bool):
            return None

        if isinstance(value, str):
            return value.strip().lstrip('"').rstrip('"')

        return value



"""vid-cleaner CLI."""

import re
from pathlib import Path
from typing import Annotated, Optional

import typer
from loguru import logger
from rich.table import Table

from vid_cleaner.__version__ import __version__
from vid_cleaner.config import Config
from vid_cleaner.constants import APP_DIR
from vid_cleaner.models import VideoFile
from vid_cleaner.utils import (
    console,
    existing_file_path,
    ffprobe,
    instantiate_logger,
    tmp_to_output,
)

typer.rich_utils.STYLE_HELPTEXT = ""

app = typer.Typer(add_completion=False, no_args_is_help=True, rich_markup_mode="rich")
config = Config(config_path=APP_DIR / "config.toml")


def version_callback(value: bool) -> None:
    """Print version and exit."""
    if value:
        console.print(f"{__package__}: v{__version__}")
        raise typer.Exit()


def parse_video_input(path: str) -> VideoFile:
    """Takes a string of a path and converts it to a VideoFile object."""
    file_path = existing_file_path(path)
    if file_path.suffix not in {".mkv", ".mp4", ".avi", "vp9", ".webm", ".mov", ".wmv"}:
        msg = f"File type '{file_path.suffix}' is not supported"
        raise typer.BadParameter(msg)

    return VideoFile(file_path)


@app.command("inspect")
def inspect_command(
    files: Annotated[
        list[VideoFile],
        typer.Argument(
            parser=parse_video_input,
            help="Path to video file(s)",
            show_default=False,
            exists=True,
            file_okay=True,
            dir_okay=True,
            resolve_path=True,
        ),
    ],
) -> None:
    """Inspect video files to display detailed stream information.

    Use this command to view detailed information about the video and audio streams
    of a video file. The information includes stream type, codec, language,
    and audio channel details. This command is useful for understanding the
    composition of a video file before performing operations like clipping or transcoding.
    """
    for video in files:
        probe = ffprobe(video.path)

        if "title" in probe["format"]["tags"]:
            name = probe["format"]["tags"]["title"]
        elif "filename" in probe["format"]:
            name = probe["format"]["filename"]
        else:
            name = video.stem

        table = Table(title=name)
        table.add_column("#")
        table.add_column("Stream")
        table.add_column("Type")
        table.add_column("Language")
        table.add_column("Channels")
        table.add_column("Channel Layout")
        table.add_column("Title")

        for i, stream in enumerate(probe["streams"]):
            language = stream["tags"].get("language", "-")
            channels = stream.get("channels", "-")
            layout = stream.get("channel_layout", "-")
            title = stream["tags"].get("title", "-")

            table.add_row(
                str(i),
                stream["codec_type"],
                stream.get("codec_name", "-"),
                language,
                str(channels),
                layout,
                title,
            )

        console.print(table)


@app.command("clip")
def clip_command(
    files: Annotated[
        list[VideoFile],
        typer.Argument(
            parser=parse_video_input,
            help="Path to video file(s)",
            show_default=False,
            exists=True,
            file_okay=True,
            dir_okay=True,
            resolve_path=True,
        ),
    ],
    start: Annotated[str, typer.Option(help="Start time 'HH:MM:SS'")] = "00:00:00",
    duration: Annotated[str, typer.Option(help="Duration to clip 'HH:MM:SS'")] = "00:01:00",
    out: Annotated[Optional[Path], typer.Option(help="Output file", show_default=False)] = None,
    overwrite: Annotated[
        bool, typer.Option("--overwrite", help="Overwrite output file if it exists")
    ] = False,
) -> None:
    """Clip a section from a video file.

    This command allows you to extract a specific portion of a video file based on start time and duration.

    * The start time and duration should be specified in [code]HH:MM:SS[/code] format.
    * You can also specify an output file to save the clipped video. If the output file is not specified, the clip will be saved with a default naming convention.

    Use the [code]--overwrite[/code] option to overwrite the output file if it already exists.
    """
    time_pattern = re.compile(r"^\d{2}:\d{2}:\d{2}$")

    if not time_pattern.match(start):
        msg = "Start must be in format HH:MM:SS"  # type: ignore [unreachable]
        raise typer.BadParameter(msg)

    if not time_pattern.match(duration):
        msg = "Duration must be in format HH:MM:SS"  # type: ignore [unreachable]
        raise typer.BadParameter(msg)

    for video in files:
        logger.info(f"⇨ {video.path.name}")

        video.clip(start, duration)
        out_file = tmp_to_output(
            video.current_tmp_file, stem=video.stem, new_file=out, overwrite=overwrite
        )
        video.cleanup()
        logger.success(f"{out_file}")


@app.command("clean")
def clean_command(
    files: Annotated[
        list[VideoFile],
        typer.Argument(
            parser=parse_video_input,
            help="Path to video file(s)",
            show_default=False,
            exists=True,
            file_okay=True,
            dir_okay=True,
            resolve_path=True,
        ),
    ],
    out: Annotated[
        Optional[Path],
        typer.Option(
            help=r"Output file [#888888]\[default: input file][/#888888]",
            show_default=False,
        ),
    ] = None,
    replace: Annotated[
        bool,
        typer.Option(
            "--replace",
            help="Delete or overwrite original file after processing. Use with caution",
        ),
    ] = False,
    downmix_stereo: Annotated[
        bool,
        typer.Option(
            "--downmix", help="Create a stereo track if none exist", rich_help_panel="Audio"
        ),
    ] = False,
    drop_original_audio: Annotated[
        bool,
        typer.Option(
            "--drop-original", help="Drop original language audio", rich_help_panel="Audio"
        ),
    ] = False,
    keep_all_subtitles: Annotated[
        bool, typer.Option("--keep-subs", help="Keep all subtitles", rich_help_panel="Subtitles")
    ] = False,
    keep_commentary: Annotated[
        bool,
        typer.Option("--keep-commentary", help="Keep commentary audio", rich_help_panel="Audio"),
    ] = False,
    keep_local_subtitles: Annotated[
        bool,
        typer.Option(
            "--keep-local-subs",
            help="Keep subtitles matching the local languages but drop all others",
            rich_help_panel="Subtitles",
        ),
    ] = False,
    subs_drop_local: Annotated[
        bool,
        typer.Option(
            "--drop-local-subs",
            help="Force dropping local subtitles even if audio is not default language",
            rich_help_panel="Subtitles",
        ),
    ] = False,
    langs: Annotated[
        str,
        typer.Option(
            help="Languages to keep. Comma separated language codes", rich_help_panel="Audio"
        ),
    ] = ",".join(config.get("keep-languages", default=["eng"])),  # type: ignore [arg-type]
    h265: Annotated[
        bool, typer.Option("--h265", help="Convert to H265", rich_help_panel="Video")
    ] = False,
    vp9: Annotated[
        bool, typer.Option("--vp9", help="Convert to VP9", rich_help_panel="Video")
    ] = False,
    force: Annotated[
        bool,
        typer.Option(
            "--force",
            help="Force processing of file even if it is transcoded",
            rich_help_panel="Video",
        ),
    ] = False,
) -> None:
    """Transcode video files to different formats or configurations.

    This command is versatile and allows for a range of transcoding options for video files with various options. You can select various audio and video settings, manage subtitles, and choose the output file format.

    The defaults for this command will:

    * Use English as the default language
    * Drop commentary audio tracks
    * Keep default language audio
    * Keep original audio if it is not the default language
    * Drop all subtitles unless the original audio is not in the default language, in which case the default subtitles are retained

    The defaults can be overridden by using the various options available.

    [bold underline]Usage Examples[/bold underline]

    [#999999]Transcode a video to H265 format and keep English audio:[/#999999]
    transcode --h265 --langs=eng <video_file>

    [#999999]Downmix audio to stereo and keep all subtitles:[/#999999]
    transcode --downmix --keep-subs <video_file>

    [#999999]Keep all audio and keep subtitles only if original audio is not English:[/#999999]
    vid-cleaner clean --local-when-needed --langs=eng <video_file>
    """
    for video in files:
        logger.info(f"⇨ {video.path.name}")

        if h265 and vp9:
            msg = "Cannot convert to both H265 and VP9"
            raise typer.BadParameter(msg)

        video.reorder_streams()

        video.process_streams(
            langs_to_keep=langs.split(","),
            drop_original_audio=drop_original_audio,
            keep_commentary=keep_commentary,
            downmix_stereo=downmix_stereo,
            keep_all_subtitles=keep_all_subtitles,
            keep_local_subtitles=keep_local_subtitles,
            subs_drop_local=subs_drop_local,
        )

        if h265:
            video._convert_to_h265(force=force)

        if vp9:
            video._convert_to_vp9(force=force)

        out_file = tmp_to_output(
            video.current_tmp_file, stem=video.stem, new_file=out, overwrite=replace
        )
        video.cleanup()

        if replace and out_file != video.path:
            logger.debug(f"Delete: {video.path}")
            video.path.unlink()

        logger.success(f"{out_file}")


@app.callback()
def main(
    log_file: Path = typer.Option(
        config.get("log_file", default=f"{APP_DIR}/vid-cleaner.log"),
        help="Path to log file",
        show_default=True,
        dir_okay=False,
        file_okay=True,
        exists=False,
    ),
    log_to_file: bool = typer.Option(
        config.get("log_to_file", default=False),
        "--log-to-file",
        help="Log to file",
        show_default=True,
    ),
    verbosity: int = typer.Option(
        0,
        "-v",
        "--verbose",
        show_default=True,
        help="""Set verbosity level(0=INFO, 1=DEBUG, 2=TRACE)""",
        count=True,
    ),
    version: Optional[bool] = typer.Option(  # noqa: ARG001
        None, "--version", help="Print version and exit", callback=version_callback, is_eager=True
    ),
) -> None:
    """Transcode video files to different formats or configurations using ffmpeg. This script provides a simple CLI for common video transcoding tasks.

    \b
    - [bold]Inspect[/bold] video files to display detailed stream information
    - [bold]Clip[/bold] a section from a video file
    - [bold]Drop audio streams[/bold] containing undesired languages or commentary
    - [bold]Drop subtitles[/bold] containing undesired languages
    - [bold]Keep subtitles[/bold] if original audio is not in desired language
    - [bold]Downmix audio[/bold] to stereo
    - [bold]Convert[/bold] video files to H265 or VP9

    [bold underline]Usage Examples[/bold underline]

        [#999999]Inspect video file:[/#999999]
        vid-cleaner inspect <video_file>

        [#999999]Clip a one minute clip from a video file:[/#999999]
        vid-cleaner clip --start=00:00:00 --duration=00:01:00 <video_file>

        [#999999]Transcode a video to H265 format and keep English audio:[/#999999]
        vid-cleaner clean --h265 --langs=eng <video_file>

        [#999999]Downmix audio to stereo and keep all subtitles:[/#999999]
        vid-cleaner clean --downmix --keep-subs <video_file>

        [#999999]Keep all audio and keep subtitles only if original audio is not English:[/#999999]
        vid-cleaner clean --local-when-needed --langs=eng <video_file>



    """  # noqa: D301
    # Instantiate Logging
    instantiate_logger(verbosity, log_file, log_to_file)


if __name__ == "__main__":
    app()



"""Constants for vid-cleaner."""

from enum import Enum
from pathlib import Path

import typer


class LogLevel(Enum):
    """Log levels for vid-cleaner."""

    INFO = 0
    DEBUG = 1
    TRACE = 2
    WARNING = 3
    ERROR = 4


class AudioLayout(Enum):
    """Audio layouts for vid-cleaner. Values are the number of streams."""

    MONO = 1
    STEREO = 2
    SURROUND5 = 6
    SURROUND7 = 8


APP_DIR = Path(typer.get_app_dir("vid-cleaner"))
EXCLUDED_VIDEO_CODECS = {"mjpeg", "mjpg", "png"}
FFMPEG_APPEND: list[str] = ["-max_muxing_queue_size", "9999"]
FFMPEG_PREPEND: list[str] = ["-y", "-hide_banner"]
H265_CODECS = {"hevc", "vp9"}

# how many bytes to read at once?
# shutil.copy uses 1024 * 1024 if _WINDOWS else 64 * 1024
# however, in my testing on MacOS with SSD, I've found a much larger buffer is faster
BUFFER_SIZE = 4096 * 1024



"""Logging utilities for vid-cleaner."""

import logging
import sys
from pathlib import Path

from loguru import logger

from vid_cleaner.constants import LogLevel


def instantiate_logger(
    verbosity: int, log_file: Path, log_to_file: bool
) -> None:  # pragma: no cover
    """Instantiate the Loguru logger for vid-cleaner.

    This function initializes the Loguru logger for the vid-cleaner application.
    It configures the logger with the specified verbosity level, log file path,
    and whether to log to a file.

    Args:
        verbosity (int): The verbosity level of the logger. Valid values are:
            - 0: No log messages will be displayed.
            - 1: Only log messages with severity level INFO and above will be displayed.
            - 2: Only log messages with severity level DEBUG and above will be displayed.
            - 3: Only log messages with severity level TRACE and above will be displayed.
        log_file (Path): The path to the log file where the log messages will be written.
        log_to_file (bool): Whether to log the messages to the file specified by `log_file`.

    Returns:
        None
    """
    logger.remove()
    logger.add(
        sys.stdout,
        level=LogLevel(verbosity).name,
        colorize=True,
        format="<level>{level: <8}</level> | <level>{message}</level> <fg #c5c5c5>({name}:{function}:{line})</fg #c5c5c5>"
        if LogLevel(verbosity) in {LogLevel.DEBUG, LogLevel.TRACE}
        else "<level>{level: <8}</level> | <level>{message}</level>",
        enqueue=True,
    )
    if log_to_file:
        logger.add(
            log_file,
            level=LogLevel(verbosity).name,
            format="{time:YYYY-MM-DD HH:mm:ss} | {level: <8} | {message} ({name})",
            rotation="50 MB",
            retention=2,
            compression="zip",
            enqueue=True,
        )

    # Intercept standard sh logs and redirect to Loguru
    if LogLevel(verbosity) in {LogLevel.DEBUG, LogLevel.TRACE}:
        logging.getLogger("sh").setLevel(level="INFO")
        logging.basicConfig(handlers=[InterceptHandler()], level=0, force=True)


class InterceptHandler(logging.Handler):  # pragma: no cover
    """Intercepts standard logging and redirects to Loguru.

    This class is a logging handler that intercepts standard logging messages and redirects them to Loguru, a third-party logging library. When a logging message is emitted, this handler determines the corresponding Loguru level for the message and logs it using the Loguru logger.

    Methods:
        emit: Intercepts standard logging and redirects to Loguru.

    Examples:
    To use the InterceptHandler with the Python logging module:
    ```
    import logging
    from logging import StreamHandler

    from loguru import logger

    # Create a new InterceptHandler and add it to the Python logging module.
    intercept_handler = InterceptHandler()
    logging.basicConfig(handlers=[StreamHandler(), intercept_handler], level=logging.INFO)

    # Log a message using the Python logging module.
    logging.info("This message will be intercepted by the InterceptHandler and logged using Loguru.")
    ```
    """

    @staticmethod
    def emit(record):  # type: ignore [no-untyped-def]
        """Intercepts standard logging and redirects to Loguru.

        This method is called by the Python logging module when a logging message is emitted. It intercepts the message and redirects it to Loguru, a third-party logging library. The method determines the corresponding Loguru level for the message and logs it using the Loguru logger.

        Args:
            record: A logging.LogRecord object representing the logging message.

        """
        # Get corresponding Loguru level if it exists.
        try:
            level = logger.level(record.levelname).name
        except ValueError:
            level = record.levelno

        # Find caller from where originated the logged message.
        frame, depth = sys._getframe(6), 6
        while frame and frame.f_code.co_filename == logging.__file__:
            frame = frame.f_back
            depth += 1

        logger.opt(depth=depth, exception=record.exc_info).log(level, record.getMessage())



"""Helper functions for vid-cleaner."""

import io
import shutil
from collections.abc import Callable
from pathlib import Path

import ffmpeg as python_ffmpeg
import requests
import typer
from loguru import logger
from rich.progress import Progress

from vid_cleaner.config import Config
from vid_cleaner.constants import APP_DIR, BUFFER_SIZE
from vid_cleaner.utils import errors


def existing_file_path(path: str) -> Path:
    """Check if the given path exists and is a file.

    Args:
        path (str): The path to check.

    Returns:
        Path: The resolved path if it exists and is a file.

    Raises:
        typer.BadParameter: If the path does not exist or is not a file.
    """
    resolved_path = Path(path).expanduser().resolve()

    if not resolved_path.exists():
        msg = f"File {path!s} does not exist"
        raise typer.BadParameter(msg)

    if not resolved_path.is_file():
        msg = f"{path!s} is not a file"
        raise typer.BadParameter(msg)

    return resolved_path


def ffprobe(path: Path) -> dict:  # pragma: no cover
    """Probe video file and return a dict.

    Args:
        path (Path): Path to video file

    Returns:
        dict: A dictionary containing information about the video file.
    """
    try:
        probe = python_ffmpeg.probe(path)
    except python_ffmpeg.Error as e:
        logger.error(e.stderr)
        raise typer.Exit(1) from e

    return probe


def query_tmdb(search: str) -> dict:  # pragma: no cover
    """Query The Movie Database API for a movie title.

    Args:
        search (str): IMDB id (tt____) to search for
        api_key (str): The Movie Database API key

    Returns:
        dict: The Movie Database API response
    """
    config = Config(config_path=APP_DIR / "config.toml")
    if not config.get("tmdb_api_key", pass_none=True):
        return {}

    url = f"https://api.themoviedb.org/3/find/{search}"

    params = {
        "api_key": config.get("tmdb_api_key"),
        "language": "en-US",
        "external_source": "imdb_id",
    }

    try:
        response = requests.get(url, params=params, timeout=15)
    except Exception as e:  # noqa: BLE001
        logger.error(e)
        return {}

    if response.status_code != 200:  # noqa: PLR2004
        logger.error(
            f"Error querying The Movie Database API: {response.status_code} {response.reason}"
        )
        return {}

    logger.trace("TMDB: Response received")
    return response.json()


def query_radarr(search: str) -> dict:  # pragma: no cover
    """Query Radarr API for a movie title.

    Args:
        search (str): Movie title to search for
        api_key (str): Radarr API key

    Returns:
        dict: Radarr API response
    """
    config = Config(config_path=APP_DIR / "config.toml")

    if not config.get("radarr_api_key", pass_none=True):
        return {}
    if not config.get("radarr_url", pass_none=True):
        return {}

    radarr_url = config.get("radarr_url")
    url = f"{radarr_url}/api/v3/parse"
    params = {
        "apikey": config.get("radarr_api_key"),
        "title": search,
    }

    try:
        response = requests.get(url, params=params, timeout=15)
    except Exception as e:  # noqa: BLE001
        logger.error(e)
        return {}

    if response.status_code != 200:  # noqa: PLR2004
        logger.error(f"Error querying Radarr: {response.status_code} {response.reason}")
        return {}

    return response.json()


def query_sonarr(search: str) -> dict:  # pragma: no cover
    """Query Sonarr API for a movie title.

    Args:
        search (str): Movie title to search for
        api_key (str): Radarr API key

    Returns:
        dict: Sonarr API response
    """
    config = Config(config_path=APP_DIR / "config.toml")

    if not config.get("sonarr_api_key", pass_none=True):
        return {}
    if not config.get("sonarr_url", pass_none=True):
        return {}

    sonarr_url = config.get("sonarr_url")
    url = f"{sonarr_url}/api/v3/parse"
    params = {
        "apikey": config.get("sonarr_api_key"),
        "title": search,
    }

    try:
        response = requests.get(url, params=params, timeout=15)
    except Exception as e:  # noqa: BLE001
        logger.error(e)
        return {}

    if response.status_code != 200:  # noqa: PLR2004
        logger.error(f"Error querying Sonarr: {response.status_code} {response.reason}")
        return {}

    logger.trace("SONARR: Response received")
    return response.json()


def _copyfileobj(
    src_bytes: io.BufferedReader,
    dest_bytes: io.BufferedWriter,
    callback: Callable,
    length: int,
) -> None:
    """Copy from src_bytes to dest_bytes.

    Args:
        src_bytes (io.BufferedReader): Source file
        dest_bytes (io.BufferedWriter): Destination file
        callback (Callable): Callback to call after every length bytes copied
        total (int): Total number of bytes to copy
        length (int): How many bytes to copy at once

    """
    copied = 0
    while True:
        buf = src_bytes.read(length)
        if not buf:
            break
        dest_bytes.write(buf)
        copied += len(buf)
        if callback is not None:
            callback(copied)


def copy_with_callback(
    src: Path,
    dest: Path,
    callback: Callable | None = None,
    buffer_size: int = BUFFER_SIZE,
) -> Path:
    """Copy file with a callback.

    Args:
        src (Path): Path to source file
        dest (Path): Path to destination file
        callback (Callable, optional): Callable callback that will be called after every buffer_size bytes copied. Defaults to None.
        buffer_size (int, optional): How many bytes to copy at once (between calls to callback). Defaults to BUFFER_SIZE (4mb).

    Returns:
        Path: Path to destination file


    Raises:
        FileNotFoundError: If source file does not exist
        SameFileError: If source and destination are the same file
        ValueError: If callback is not callable

    Note: Does not copy extended attributes, resource forks or other metadata.
    """
    if not src.is_file():
        msg = f"src file `{src}` doesn't exist"
        raise FileNotFoundError(msg)

    dest = dest / src.name if dest.is_dir() else dest

    if dest.exists() and src.samefile(dest):
        msg = f"source file `{src}` and destination file `{dest}` are the same file."
        raise errors.SameFileError(msg)

    if callback is not None and not callable(callback):
        msg = f"callback must be callable, not {type(callback)}"  # type: ignore [unreachable]
        raise ValueError(msg)

    with src.open("rb") as src_bytes, dest.open("wb") as dest_bytes:
        _copyfileobj(src_bytes, dest_bytes, callback=callback, length=buffer_size)

    shutil.copymode(str(src), str(dest))

    return dest


def tmp_to_output(
    tmp_file: Path,
    stem: str,
    overwrite: bool = False,
    new_file: Path | None = None,
) -> Path:
    """Copy a temporary file to an output file.

    If no output file is given, the name and directory of the source file will be used.  If overwrite is False, a number will be appended to the stem if the output file already exists.

    Args:
        tmp_file (Path): Path to input file
        stem (str): Stem of output file
        new_file (Path, optional): Path to output file. Defaults to None.
        overwrite (bool, optional): Overwrite output file if it exists. Defaults to False.

    Returns:
        Path: Path to output file
    """
    # When a path is given, use that
    if new_file:
        parent = new_file.parent.expanduser().resolve()
        stem = new_file.stem
    else:
        parent = Path.cwd()

    # Ensure parent directory exists
    parent.mkdir(parents=True, exist_ok=True)

    new = parent / f"{stem}{tmp_file.suffix}"

    if not overwrite:
        i = 1
        while new.exists():
            new = parent / f"{stem}_{i}{tmp_file.suffix}"
            i += 1

    tmp_file_size = tmp_file.stat().st_size

    with Progress(transient=True) as progress:
        task = progress.add_task("Copy file…", total=tmp_file_size)
        copy_with_callback(
            tmp_file,
            new,
            callback=lambda total_copied: progress.update(task, completed=total_copied),
        )

    logger.trace(f"File copied to {new}")
    return new
